import dataclasses
import astropy.units as u
import named_arrays as na
import optika

__all__ = [
    "AbstractRulingSpacing",
    "ConstantRulingSpacing",
    "Polynomial1dRulingSpacing",
    "HolographicRulingSpacing",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractRulingSpacing(
    optika.mixins.Printable,
    optika.mixins.Transformable,
    optika.mixins.Shaped,
):
    """
    An interface describing the instantaneous ruling spacing on the surface
    of a diffraction grating.

    This is useful if you want to define a grating with variable line spacing.
    """

    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractCartesian3dVectorArray:
        """
        The local ruling vector at the given position

        Parameters
        ----------
        position
            The location on the grating at which to evaluate the ruling spacing.
        normal
            The unit vector perpendicular to the surface at the given position.
        """


@dataclasses.dataclass(eq=False, repr=False)
class ConstantRulingSpacing(
    AbstractRulingSpacing,
):
    """
    The simplest type of ruling spacing, a constant distance between each ruling.
    """

    constant: u.Quantity | na.AbstractScalar
    """The constant describing the ruling spacing."""

    normal: na.AbstractCartesian3dVectorArray
    """The unit vector normal to the planes of the rulings."""

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.constant),
            optika.shape(self.normal),
        )

    @property
    def transformation(self) -> None:
        return None

    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.Cartesian3dVectorArray:
        return self.constant * self.normal


@dataclasses.dataclass(eq=False, repr=False)
class Polynomial1dRulingSpacing(
    AbstractRulingSpacing,
):
    """
    Ruling spacing specified by a 1-dimensional polynomial.
    """

    coefficients: dict[int, u.Quantity | na.AbstractScalar]
    """
    The coefficients of the polynomial represented as a dictionary where
    the values are the coefficients and the keys are the power associated
    with each coefficient.
    """

    normal: na.AbstractCartesian3dVectorArray
    """The unit vector normal to the planes of the rulings."""

    transformation: None | na.transformations.AbstractTransformation = None
    """
    An arbitrary coordinate system transformation applied to the argument
    of the polynomial.
    """

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.coefficients),
            optika.shape(self.normal),
            optika.shape(self.transformation),
        )

    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.Cartesian3dVectorArray:

        coefficients = self.coefficients
        normal_rulings = self.normal
        transformation = self.transformation

        if transformation is not None:
            position = transformation(position)

        x = position @ normal_rulings

        result = 0 * u.mm
        for power, coefficient in coefficients.items():
            result = result + coefficient * (x**power)

        return result * normal_rulings


@dataclasses.dataclass(eq=False, repr=False)
class HolographicRulingSpacing(
    AbstractRulingSpacing,
):
    r"""
    Rulings created by interfering two beams.

    Examples
    --------

    Create some holographic rulings from two source points,
    launch rays from the first source point and confirm they are refocused
    onto the second source point.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        # Define the origins of the two recording beams
        x1 = na.Cartesian3dVectorArray(5, 0, -10) * u.mm
        x2 = na.Cartesian3dVectorArray(10, 0, -10) * u.mm

        # Define the wavelength of the recording beams
        wavelength = 500 * u.nm

        # Define the surface normal
        normal = na.Cartesian3dVectorArray(0, 0, -1)

        # Define input rays emanating from the origin
        # of the first recording beam
        position = na.Cartesian3dVectorArray(
            x=na.linspace(-5, +5, axis="x", num=5) * u.mm,
            y=0 * u.mm,
            z=0 * u.mm,
        )
        direction_input = position - x1

        # Initialize the holographic ruling spacing
        # representation
        rulings = optika.rulings.HolographicRulingSpacing(
            x1=x1,
            x2=x2,
            wavelength=wavelength,
        )

        # Evaluate the ruling spacing where
        # the rays strike the surface
        d = rulings(position, normal)

        # Compute the new direction of some diffracted
        # rays
        direction_output = optika.materials.snells_law(
            wavelength=wavelength,
            direction=direction_input.normalized,
            index_refraction=1,
            index_refraction_new=1,
            normal=normal,
            is_mirror=True,
            diffraction_order=1,
            spacing_rulings=d.length,
            normal_rulings=d.normalized,
        )
        direction_output = direction_output * 20 * u.mm

        # Plot the results
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots()
            na.plt.plot(
                na.stack([x1, position], axis="t"),
                components=("z", "x"),
                axis="t",
                color="tab:blue",
            )
            na.plt.plot(
                na.stack([position, position + direction_output], axis="t"),
                components=("z", "x"),
                axis="t",
                color="tab:orange",
            )
            ax.axvline(0)
            ax.scatter(x1.z, x1.x)
            ax.scatter(x2.z, x2.x)

    Notes
    -----

    From :cite:t:`Welford1975`, the ruling spacing is given by

    .. math::

        \mathbf{d} = \frac{\lambda}{a} \mathbf{q} \times \mathbf{n}

    where :math:`\lambda` is the wavelength of the recording beams,
    :math:`\mathbf{n}` is a unit vector perpendicular to the surface,

    .. math::

        a \mathbf{q} = \mathbf{n} \times (\pm \mathbf{r}_1 \mp \mathbf{r}_2),

    :math:`\mathbf{r}_1` is a unit vector in the direction of the first
    recording beam,
    and :math:`\mathbf{r}_2` is a unit vector in the direction of the second
    recording beam.
    If rays are diverging from the origin of the recording beams,
    the top branch is used, otherwise the bottom branch is used.
    """

    x1: na.AbstractCartesian3dVectorArray
    """
    The origin of the first recording beam in local coordinates.
    """

    x2: na.AbstractCartesian3dVectorArray
    """
    The origin of the second recording beam in local coordinates.
    """

    wavelength: u.Quantity | na.AbstractScalar
    """
    The wavelength of the recording beams.
    """

    is_diverging_1: bool | na.AbstractScalar = True
    """
    A boolean flag indicating if rays are diverging from the origin of the
    first beam.
    """

    is_diverging_2: bool | na.AbstractScalar = False
    """
    A boolean flag indicating if rays are diverging from the origin of the
    second beam.
    """

    transformation: None | na.transformations.AbstractTransformation = None
    """
    A transformation from surface coordinates to ruling coordinates.
    """

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.x1),
            optika.shape(self.x2),
            optika.shape(self.is_diverging_1),
            optika.shape(self.is_diverging_2),
        )

    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.Cartesian3dVectorArray:

        x1 = self.x1
        x2 = self.x2
        wavelength = self.wavelength
        d1 = self.is_diverging_1
        d2 = self.is_diverging_2
        n = normal

        d1 = 2 * d1 - 1
        d2 = 2 * d2 - 1

        r1 = position - x1
        r2 = position - x2

        r1 = d1 * r1.normalized
        r2 = d2 * r2.normalized

        dr = r1 - r2

        aq = n.cross(dr)

        a = aq.length
        q = aq / a

        spacing = wavelength / a

        result = spacing * q.cross(n)

        return result



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
from . import test_mixins


class AbstractTestAbstractSystem(
    test_mixins.AbstractTestPlottable,
    test_mixins.AbstractTestPrintable,
    test_mixins.AbstractTestTransformable,
    test_mixins.AbstractTestShaped,
):

    @pytest.mark.parametrize(
        argnames="scene",
        argvalues=[
            na.FunctionArray(
                inputs=na.SpectralPositionalVectorArray(
                    wavelength=na.linspace(
                        start=530 * u.nm,
                        stop=531 * u.nm,
                        axis="wavelength",
                        num=2,
                    ),
                    position=na.Cartesian2dVectorLinearSpace(
                        start=-1,
                        stop=+1,
                        axis=na.Cartesian2dVectorArray("field_x", "field_y"),
                        num=11,
                    ),
                ),
                outputs=na.random.uniform(
                    low=0 * u.photon / u.cm**2 / u.arcsec**2 / u.s / u.nm,
                    high=100 * u.photon / u.cm**2 / u.arcsec**2 / u.s / u.nm,
                    shape_random=dict(field_x=10, field_y=10),
                ),
            )
        ],
    )
    def test_image(
        self,
        a: optika.systems.AbstractSystem,
        scene: na.FunctionArray[na.SpectralPositionalVectorArray, na.AbstractScalar],
    ):
        result = a.image(scene)
        assert isinstance(result, na.FunctionArray)
        assert isinstance(result.inputs, na.SpectralPositionalVectorArray)
        assert isinstance(result.outputs, na.AbstractScalar)
        assert np.all(result.inputs.wavelength > 0 * u.nm)
        assert na.unit_normalized(result.inputs.position).is_equivalent(u.mm)
        assert result.outputs.sum() > (0 * u.electron)


class AbstractTestAbstractSequentialSystem(
    AbstractTestAbstractSystem,
):
    def test_object(self, a: optika.systems.AbstractSequentialSystem):
        if a.object is not None:
            assert isinstance(a.object, optika.surfaces.AbstractSurface)

    def test_object_is_at_infinity(self, a: optika.systems.AbstractSequentialSystem):
        result = a.object_is_at_infinity
        assert isinstance(result, bool)

    def test_surfaces(self, a: optika.systems.AbstractSequentialSystem):
        for surface in a.surfaces:
            assert isinstance(surface, optika.surfaces.AbstractSurface)

    def test_sensor(self, a: optika.systems.AbstractSequentialSystem):
        if a.sensor is not None:
            assert isinstance(a.sensor, optika.sensors.AbstractImagingSensor)

    def test_axis_surface(self, a: optika.systems.AbstractSequentialSystem):
        assert isinstance(a.axis_surface, str)

    def test_surfaces_all(self, a: optika.systems.AbstractSequentialSystem):
        for surface in a.surfaces_all:
            assert isinstance(surface, optika.surfaces.AbstractSurface)

    def test_grid_input(self, a: optika.systems.AbstractSequentialSystem):
        assert isinstance(a.grid_input, optika.vectors.ObjectVectorArray)

    def test_index_field_stop(self, a: optika.systems.AbstractSequentialSystem):
        assert isinstance(a.index_field_stop, int)
        assert a.surfaces_all[a.index_field_stop].is_field_stop

    def test_index_pupil_stop(self, a: optika.systems.AbstractSequentialSystem):
        assert isinstance(a.index_pupil_stop, int)
        assert a.surfaces_all[a.index_pupil_stop].is_pupil_stop

    def test_field_stop(self, a: optika.systems.AbstractSequentialSystem):
        assert a.field_stop.is_field_stop

    def test_pupil_stop(self, a: optika.systems.AbstractSequentialSystem):
        assert a.pupil_stop.is_pupil_stop

    def test_rayfunction_stops(self, a: optika.systems.AbstractSequentialSystem):
        result = a.rayfunction_stops
        assert isinstance(result, optika.rays.RayFunctionArray)
        assert isinstance(result.inputs, optika.vectors.ObjectVectorArray)
        assert isinstance(result.outputs, optika.rays.RayVectorArray)
        assert result.ndim >= 2

    def test_field_min(self, a: optika.systems.AbstractSequentialSystem):
        result = a.field_min
        assert isinstance(result, na.AbstractCartesian2dVectorArray)
        if a.object_is_at_infinity:
            assert na.unit(result).is_equivalent(u.deg)
        else:
            assert na.unit(result).is_equivalent(u.m)

    def test_field_max(self, a: optika.systems.AbstractSequentialSystem):
        result = a.field_max
        assert isinstance(result, na.AbstractCartesian2dVectorArray)
        assert np.all(result > a.field_min)
        if a.object_is_at_infinity:
            assert na.unit(result).is_equivalent(u.deg)
        else:
            assert na.unit(result).is_equivalent(u.m)

    def test_pupil_min(self, a: optika.systems.AbstractSequentialSystem):
        result = a.pupil_min
        assert isinstance(result, na.AbstractCartesian2dVectorArray)
        if a.object_is_at_infinity:
            assert na.unit(result).is_equivalent(u.m)
        else:
            assert na.unit(result).is_equivalent(u.deg)

    def test_pupil_max(self, a: optika.systems.AbstractSequentialSystem):
        result = a.pupil_max
        assert isinstance(result, na.AbstractCartesian2dVectorArray)
        assert np.all(result > a.pupil_min)
        if a.object_is_at_infinity:
            assert na.unit(result).is_equivalent(u.m)
        else:
            assert na.unit(result).is_equivalent(u.deg)

    @pytest.mark.parametrize(
        argnames="wavelength,field,pupil",
        argvalues=[
            (
                None,
                None,
                None,
            ),
            (
                500 * u.nm,
                na.Cartesian2dVectorLinearSpace(
                    start=-1,
                    stop=1,
                    axis=na.Cartesian2dVectorArray("x", "y"),
                    num=11,
                ),
                na.Cartesian2dVectorLinearSpace(
                    start=-1,
                    stop=1,
                    axis=na.Cartesian2dVectorArray("x", "y"),
                    num=11,
                ),
            ),
        ],
    )
    def test_raytrace(
        self,
        a: optika.systems.AbstractSequentialSystem,
        wavelength: None | u.Quantity | na.AbstractScalar,
        field: None | na.AbstractCartesian2dVectorArray,
        pupil: None | na.AbstractCartesian2dVectorArray,
    ):
        raytrace = a.raytrace(
            wavelength=wavelength,
            field=field,
            pupil=pupil,
        )
        assert isinstance(raytrace, optika.rays.RayFunctionArray)
        assert isinstance(raytrace.inputs, optika.vectors.ObjectVectorArray)
        assert isinstance(raytrace.outputs, optika.rays.RayVectorArray)
        assert a.axis_surface in raytrace.shape

    @pytest.mark.parametrize(
        argnames="wavelength,field,pupil",
        argvalues=[
            (
                None,
                None,
                None,
            ),
            (
                500 * u.nm,
                na.Cartesian2dVectorLinearSpace(
                    start=-1,
                    stop=1,
                    axis=na.Cartesian2dVectorArray("x", "y"),
                    num=11,
                ),
                na.Cartesian2dVectorLinearSpace(
                    start=-1,
                    stop=1,
                    axis=na.Cartesian2dVectorArray("x", "y"),
                    num=11,
                ),
            ),
        ],
    )
    def test_rayfunction(
        self,
        a: optika.systems.AbstractSequentialSystem,
        wavelength: None | u.Quantity | na.AbstractScalar,
        field: None | na.AbstractCartesian2dVectorArray,
        pupil: None | na.AbstractCartesian2dVectorArray,
    ):
        raytrace = a.rayfunction(
            wavelength=wavelength,
            field=field,
            pupil=pupil,
        )
        assert isinstance(raytrace, optika.rays.RayFunctionArray)
        assert isinstance(raytrace.inputs, optika.vectors.ObjectVectorArray)
        assert isinstance(raytrace.outputs, optika.rays.RayVectorArray)
        assert a.axis_surface not in raytrace.shape

    def test_rayfunction_default(self, a: optika.systems.AbstractSequentialSystem):
        rayfunction = a.rayfunction_default
        assert isinstance(rayfunction, optika.rays.RayFunctionArray)
        assert isinstance(rayfunction.inputs, optika.vectors.ObjectVectorArray)
        assert isinstance(rayfunction.outputs, optika.rays.RayVectorArray)
        assert a.axis_surface not in rayfunction.shape


_objects = [
    None,
    optika.surfaces.Surface(),
    optika.surfaces.Surface(
        aperture=optika.apertures.CircularAperture(10 * u.mm),
    ),
    optika.surfaces.Surface(
        aperture=optika.apertures.CircularAperture(0.1),
    ),
]

_surfaces = [
    optika.surfaces.Surface(
        name="mirror",
        sag=optika.sags.SphericalSag(-200 * u.mm),
        material=optika.materials.Mirror(),
        aperture=optika.apertures.CircularAperture(20 * u.mm),
        is_pupil_stop=True,
        transformation=na.transformations.Cartesian3dTranslation(z=100 * u.mm),
    ),
]

_sensor = optika.sensors.ImagingSensor(
    name="sensor",
    width_pixel=15 * u.um,
    axis_pixel=na.Cartesian2dVectorArray("detector_x", "detector_y"),
    timedelta_exposure=1 * u.s,
    num_pixel=na.Cartesian2dVectorArray(128, 128),
    transformation=na.transformations.Cartesian3dTranslation(z=1 * u.mm),
    is_field_stop=True,
)

_grid_input = optika.vectors.ObjectVectorArray(
    wavelength=500 * u.nm,
    field=na.Cartesian2dVectorLinearSpace(
        start=0,
        stop=1,
        axis=na.Cartesian2dVectorArray("field_x", "field_y"),
        num=5,
    ),
    pupil=na.Cartesian2dVectorLinearSpace(
        start=0,
        stop=1,
        axis=na.Cartesian2dVectorArray("pupil_x", "pupil_y"),
        num=5,
    ),
)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.systems.SequentialSystem(
            object=obj,
            surfaces=_surfaces,
            sensor=_sensor,
            grid_input=_grid_input,
        )
        for obj in _objects
    ],
)
class TestSequentialSystem(AbstractTestAbstractSequentialSystem):
    pass



import pytest
import optika
from .._depletion_test import AbstractTestAbstractJanesickDepletionModel


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.sensors.E2VCCD64ThinDepletionModel(),
    ],
)
class TestE2VCCD64ThinDepletionModel(
    AbstractTestAbstractJanesickDepletionModel,
):
    pass



import pathlib
import dataclasses
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
from .._depletion import AbstractJanesickDepletionModel

__all__ = [
    "E2VCCD64ThinDepletionModel",
]


@dataclasses.dataclass(eq=False, repr=False)
class E2VCCD64ThinDepletionModel(
    AbstractJanesickDepletionModel,
):
    """
    A model of the depletion region of a "thin" (20 :math:`\Omega`-cm)
    e2v CCD64 imaging sensor, which uses charge diffusion measurements from
    :cite:t:`Stern2004` to estimate the thickness of the depletion region.

    Examples
    --------

    Plot the measured e2v CCD64 measured mean charge capture vs. the fitted
    mean charge capture.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        # Create a new instance of the e2v CCD64 depletion region model
        depletion = optika.sensors.E2VCCD64ThinDepletionModel()

        # Store the wavelengths at which the MCC was measured
        wavelength_measured = depletion.mean_charge_capture_measured.inputs

        # Store the MCC measurements
        mcc_measured = depletion.mean_charge_capture_measured.outputs

        # Define a grid of wavelengths with which to evaluate the fitted MCC
        wavelength_fit = na.geomspace(1, 10000, axis="wavelength", num=1001) * u.AA
        energy_fit = wavelength_fit.to(u.eV, equivalencies=u.spectral())

        # Evaluate the fitted MCC using the given wavelengths
        mcc_fit = depletion.mean_charge_capture(wavelength_fit)

        # Plot the measured QE vs the fitted QE
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots(constrained_layout=True)
            ax2 = ax.twiny()
            ax2.invert_xaxis()
            na.plt.scatter(
                wavelength_measured,
                mcc_measured,
                ax=ax,
                label="measured",
            )
            na.plt.plot(
                wavelength_fit,
                mcc_fit,
                ax=ax,
                label="fit",
            )
            na.plt.plot(
                energy_fit,
                mcc_fit,
                ax=ax2,
                linestyle="None",
            )
            ax.set_xscale("log")
            ax2.set_xscale("log")
            ax.set_xlabel(f"wavelength ({ax.get_xlabel()})")
            ax2.set_xlabel(f"energy ({ax2.get_xlabel()})")
            ax.set_ylabel("mean charge capture")
            ax.legend()

    The thickness of the depletion region found by the fit is

    .. jupyter-execute::

        depletion.thickness
    """

    @property
    def shape(self) -> dict[str, int]:
        return dict()

    @property
    def chemical_substrate(self) -> optika.chemicals.Chemical:
        return optika.chemicals.Chemical("Si")

    @property
    def thickness_substrate(self) -> u.Quantity:
        return 8 * u.um

    @property
    def width_pixel(self) -> u.Quantity:
        return 16 * u.um

    @property
    def mean_charge_capture_measured(self) -> na.FunctionArray:
        directory = pathlib.Path(__file__).parent

        energy, mcc = np.genfromtxt(
            fname=directory / "_e2v_ccd64_thin_stern2004.csv",
            delimiter=", ",
            unpack=True,
        )
        energy = energy << u.keV
        wavelength = energy.to(u.AA, equivalencies=u.spectral())

        return na.FunctionArray(
            inputs=na.ScalarArray(wavelength, axes="wavelength"),
            outputs=na.ScalarArray(mcc, axes="wavelength"),
        )



import abc
import dataclasses
import astropy.units as u
import named_arrays as na
import optika

__all__ = [
    "AbstractSlopeErrorParameters",
    "SlopeErrorParameters",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractSlopeErrorParameters(
    optika.mixins.Printable,
    optika.mixins.Shaped,
):
    """collection of parameters used to compute the slope error"""

    @property
    @abc.abstractmethod
    def step_size(self) -> na.ScalarLike:
        """the horizontal distance to use when measuring the slope"""

    @property
    @abc.abstractmethod
    def kernel_size(self) -> na.ScalarLike:
        """
        size of the boxcar kernel that is convolved with the wavefront error
        before measuring the slope
        """


@dataclasses.dataclass(eq=False, repr=False)
class SlopeErrorParameters(
    AbstractSlopeErrorParameters,
):
    step_size: na.ScalarLike = 0 * u.mm
    kernel_size: na.ScalarLike = 0 * u.mm

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.step_size),
            optika.shape(self.kernel_size),
        )



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
from .._tests import test_mixins


class AbstractTestAbstractRulingSpacing(
    test_mixins.AbstractTestPrintable,
    test_mixins.AbstractTestTransformable,
    test_mixins.AbstractTestShaped,
):
    @pytest.mark.parametrize(
        argnames="position",
        argvalues=[
            na.Cartesian3dVectorArray(0, 0, 0) * u.mm,
            na.Cartesian3dVectorArray(
                x=na.linspace(-1, 1, axis="position_x", num=5) * u.mm,
                y=na.linspace(-1, 1, axis="position_y", num=6) * u.mm,
                z=0 * u.mm,
            ),
        ],
    )
    @pytest.mark.parametrize(
        argnames="normal",
        argvalues=[
            na.Cartesian3dVectorArray(0, 0, -1),
        ],
    )
    def test__call__(
        self,
        a: optika.rulings.AbstractRulingSpacing,
        position: na.AbstractCartesian3dVectorArray,
        normal: na.Cartesian3dVectorArray,
    ):
        result = a(position, normal)
        assert isinstance(result, na.AbstractCartesian3dVectorArray)
        assert np.all(result.length > (0 * u.mm))


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.rulings.ConstantRulingSpacing(
            constant=1 * u.um,
            normal=na.Cartesian3dVectorArray(1, 0, 0),
        ),
    ],
)
class TestConstantSpacing(
    AbstractTestAbstractRulingSpacing,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.rulings.Polynomial1dRulingSpacing(
            coefficients={
                0: 1 * u.um,
                1: 2,
                2: 3 / u.um,
            },
            normal=na.Cartesian3dVectorArray(1, 0, 0),
            transformation=na.transformations.Cartesian3dRotationZ(30 * u.deg),
        ),
    ],
)
class TestPolynomial1dRulingSpacing(
    AbstractTestAbstractRulingSpacing,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.rulings.HolographicRulingSpacing(
            x1=na.Cartesian3dVectorArray(0, 1, 2) * u.mm,
            x2=na.Cartesian3dVectorArray(1, 2, 3) * u.mm,
            wavelength=500 * u.nm,
        ),
    ],
)
class TestHolographicRulingSpacing(
    AbstractTestAbstractRulingSpacing,
):
    pass



import abc
import re
import pathlib
import dataclasses
import numpy as np
import astropy.units as u
import named_arrays as na
import optika

__all__ = [
    "AbstractChemical",
    "Chemical",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractChemical(
    optika.mixins.Printable,
    optika.mixins.Shaped,
):
    """
    Interface defining the optical constants for a given chemical.
    """

    @property
    @abc.abstractmethod
    def formula(self) -> str | na.AbstractScalar:
        """
        the `empirical formula <https://en.wikipedia.org/wiki/Empirical_formula>`_
        of the chemical compound.

        For example, water would be expressed as ``"H2O"``
        and hydrogen peroxide would be expressed as ``"H2O2"``.
        """

    @property
    def formula_latex(self) -> str:
        """
        LaTeX representation of the chemical formula, with appropriate subscripts.
        """
        formula = self.formula
        pattern = r"\d+"
        repl = r"$_\g<0>$"
        if isinstance(formula, na.ScalarArray):
            formula = formula.copy()
            for index in formula.ndindex():
                formula[index] = re.sub(pattern, repl, formula[index].ndarray)
        else:
            formula = re.sub(pattern, repl, formula)
        return formula

    @property
    @abc.abstractmethod
    def is_amorphous(self) -> bool:
        """
        Boolean flag controlling whether the chemical is amorphous
        or crystalline.
        """

    @property
    @abc.abstractmethod
    def table(self) -> None | str:
        """
        Name of the table of chemical constants.
        """

    @property
    def file_nk(self) -> na.ScalarArray:
        """
        The path to the :cite:t:`Windt1998` file containing the
        index of refaction, :math:`n`, and the wavenumber, :math:`k`.
        """
        formula = na.as_named_array(self.formula)
        is_amorphous = na.as_named_array(self.is_amorphous)
        table = na.as_named_array(self.table)

        shape = na.shape_broadcasted(formula, is_amorphous, table)

        formula = formula.broadcast_to(shape)
        is_amorphous = is_amorphous.broadcast_to(shape)
        table = table.broadcast_to(shape)

        result = na.ScalarArray.empty(shape=shape, dtype=pathlib.Path)

        path_base = pathlib.Path(__file__).parent / "nk"
        for index in result.ndindex():
            file = f"{formula[index].ndarray}"

            if table[index].ndarray is not None:
                file = f"{file}_{table[index].ndarray}"

            if is_amorphous[index]:
                file = f"a-{file}"

            file = f"{file}.nk"

            result[index] = path_base / file

        return result

    def n(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
    ) -> na.AbstractScalar:
        """
        The complex index of refaction of this chemical for a given wavelength
        """
        file_nk = self.file_nk
        shape_base = file_nk.shape

        shape = na.broadcast_shapes(shape_base, wavelength.shape)

        wavelength = na.broadcast_to(wavelength, shape)
        result = np.empty_like(wavelength.value, dtype=complex)

        for i, index in enumerate(file_nk.ndindex()):
            file_nk_index = file_nk[index].ndarray

            skip_header = 0
            with open(file_nk_index, "r") as f:
                for line in f:
                    if line.startswith(";"):
                        skip_header += 1
                    else:
                        break

            wavelength_index, n_index, k_index = np.loadtxt(
                fname=file_nk_index,
                skiprows=skip_header,
                unpack=True,
            )

            wavelength_index = wavelength_index << u.AA
            wavelength_index = na.ScalarArray(wavelength_index, axes="wavelength")
            n_index = na.ScalarArray(n_index, axes="wavelength")
            k_index = na.ScalarArray(k_index, axes="wavelength")

            result[index] = na.interp(
                x=wavelength[index],
                xp=wavelength_index,
                fp=n_index + 1j * k_index,
            )

        return result

    def index_refraction(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
    ) -> na.AbstractScalar:
        """
        The index of refraction of this chemical for the given wavelength.
        """
        return np.real(self.n(wavelength))

    def wavenumber(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
    ) -> na.AbstractScalar:
        """
        The wavenumber of this chemical for the given wavelength.
        """
        return np.imag(self.n(wavelength))

    def absorption(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
    ):
        """
        The absorption coefficient of this chemical for the given wavelength.

        Parameters
        ----------
        wavelength
            The wavelength of light in vacuum for which to compute the
            absorption coefficient.
        """
        return 4 * np.pi * self.wavenumber(wavelength) / wavelength


@dataclasses.dataclass(eq=False, repr=False)
class Chemical(
    AbstractChemical,
):
    """
    An object that represents the optical properties of a chemical.

    Uses the tabulated optical constants from :cite:t:`Windt1998`.

    Examples
    --------

    Plot the indices of refractions of silicon and silicon dioxide.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        si = optika.chemicals.Chemical("Si")
        sio2 = optika.chemicals.Chemical("SiO2")

        wavelength = na.geomspace(10, 10000, axis="wavelength", num=1001) * u.AA

        n_si = si.index_refraction(wavelength)
        n_sio2 = sio2.index_refraction(wavelength)

        fig, ax = plt.subplots(constrained_layout=True)
        na.plt.plot(wavelength, n_si, label="silicon");
        na.plt.plot(wavelength, n_sio2, label="silicon dioxide");
        ax.set_xscale("log");
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel("index of refraction");
        ax.legend();

    Plot the wavenumbers of silicon and silicon dioxide

    .. jupyter-execute::

        k_si = si.wavenumber(wavelength)
        k_sio2 = sio2.wavenumber(wavelength)

        fig, ax = plt.subplots(constrained_layout=True)
        na.plt.plot(wavelength, k_si, label="silicon");
        na.plt.plot(wavelength, k_sio2, label="silicon dioxide");
        ax.set_xscale("log");
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel("wavenumber");
        ax.legend();
    """

    formula: str | na.AbstractScalar = dataclasses.MISSING
    """
    the `empirical formula <https://en.wikipedia.org/wiki/Empirical_formula>`_
    of the chemical compound.

    For example, water would be expressed as ``"H2O"``
    and hydrogen peroxide would be expressed as ``"H2O2"``.
    """

    is_amorphous: bool = False
    """
    Boolean flag controlling whether the chemical is amorphous
    or crystalline.
    """

    table: None | str = None
    """
    Name of the table of chemical constants.

    Common options are ``"palik"``, ``"llnl"``, and ``"windt"``.
    The database is the same as the IMD code :cite:p:`Windt1998`.
    The default value, :obj:`None`, usually means concatenating the
    tables in :cite:t:`Palik1997` and :cite:t:`Henke1993`.
    """

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.formula),
            optika.shape(self.is_amorphous),
            optika.shape(self.table),
        )



from __future__ import annotations
import abc
import dataclasses
import named_arrays as na

__all__ = [
    "AbstractPolarizationVectorArray",
    "PolarizationVectorArray",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractPolarizationVectorArray(
    na.AbstractCartesianVectorArray,
):
    """
    An interface describing a vector in the :math:`s` and :math:`p` coordinate
    system.
    """

    @property
    @abc.abstractmethod
    def s(self) -> float | na.AbstractScalar:
        """
        The component of the electric field perpendicular to the plane of incidence.
        """

    @property
    @abc.abstractmethod
    def p(self) -> float | na.AbstractScalar:
        """
        The component of the electric field parallel to the plane of incidence.
        """

    @property
    def average(self) -> float | na.AbstractScalar:
        return (self.s + self.p) / 2

    @property
    def type_abstract(self) -> type[AbstractPolarizationVectorArray]:
        return AbstractPolarizationVectorArray

    @property
    def type_explicit(self) -> type[PolarizationVectorArray]:
        return PolarizationVectorArray

    @property
    def type_matrix(self) -> type[na.AbstractMatrixArray]:
        raise NotImplementedError

    @property
    def explicit(self) -> PolarizationVectorArray:
        return super().explicit


@dataclasses.dataclass(eq=False, repr=False)
class PolarizationVectorArray(
    AbstractPolarizationVectorArray,
    na.AbstractExplicitCartesianVectorArray,
):
    """
    A vector described in terms of components parallel and perpendicular
    to the plane of incidence.
    """

    s: float | na.AbstractScalar = 0
    """The component of the electric field perpendicular to the plane of incidence."""

    p: float | na.AbstractScalar = 0
    """The component of the electric field parallel to the plane of incidence."""

    @classmethod
    def from_scalar(
        cls,
        scalar: na.ScalarLike,
        like: None | PolarizationVectorArray = None,
    ) -> PolarizationVectorArray:
        if like is None:
            type_like = cls
        else:
            type_like = type(like)
        return type_like(
            s=scalar,
            p=scalar,
        )



import pytest
import numpy as np
import matplotlib.axes
import astropy.units as u
import named_arrays as na
import optika
import optika.rays._tests.test_ray_vectors
from . import test_mixins


active_parameterization = [
    True,
    na.linspace(-1, 1, "active", 3) >= 0,
]

inverted_parameterization = [
    False,
    na.linspace(-1, 1, "inverted", 4) >= 0,
]

transform_parameterization = [
    None,
    na.transformations.Cartesian3dRotationZ(30 * u.deg),
]


class AbstractTestAbstractAperture(
    test_mixins.AbstractTestPrintable,
    test_mixins.AbstractTestPlottable,
    test_mixins.AbstractTestTransformable,
    test_mixins.AbstractTestShaped,
):
    def test_samples_wire(self, a: optika.apertures.AbstractAperture):
        assert isinstance(a.samples_wire, int)
        assert a.samples_wire > 0

    def test_active(self, a: optika.apertures.AbstractAperture):
        assert isinstance(a.active, (bool, na.AbstractScalar))
        assert np.issubdtype(na.get_dtype(a.active), bool)

    def test_inverted(self, a: optika.apertures.AbstractAperture):
        assert isinstance(a.inverted, (bool, na.AbstractScalar))
        assert np.issubdtype(na.get_dtype(a.inverted), bool)

    def test__call__(self, a: optika.apertures.AbstractAperture):
        position = na.Cartesian3dVectorLinearSpace(
            start=a.bound_lower,
            stop=a.bound_upper,
            axis=na.Cartesian3dVectorArray("x", "y", "z"),
            num=na.Cartesian3dVectorArray(5, 5, 1),
        )

        result = a(position)
        assert np.issubdtype(na.get_dtype(result), bool)
        assert np.any(result)

    @pytest.mark.parametrize("rays", optika.rays._tests.test_ray_vectors.rays)
    def test_clip_rays(
        self,
        a: optika.apertures.AbstractAperture,
        rays: optika.rays.RayVectorArray,
    ):
        result = a.clip_rays(rays)
        assert isinstance(result, optika.rays.RayVectorArray)
        assert result is not rays
        assert np.all(result.position == rays.position)
        assert np.all(result.direction == rays.direction)
        assert np.all(result.intensity == rays.intensity)
        assert np.all(result.attenuation == rays.attenuation)
        assert np.all(result.index_refraction == rays.index_refraction)
        assert result.unvignetted is not rays.unvignetted
        assert np.mean(result.unvignetted) <= np.mean(rays.unvignetted)

    def test_bound_lower(self, a: optika.apertures.AbstractAperture):
        assert isinstance(a.bound_lower, na.AbstractCartesian3dVectorArray)

    def test_bound_upper(self, a: optika.apertures.AbstractAperture):
        assert isinstance(a.bound_upper, na.AbstractCartesian3dVectorArray)
        assert np.all(a.bound_upper.x != a.bound_lower.x)
        assert np.all(a.bound_upper.y != a.bound_lower.y)

    def test_vertices(self, a: optika.apertures.AbstractAperture):
        if a.vertices is not None:
            assert isinstance(a.vertices, na.AbstractCartesian3dVectorArray)
            assert "vertex" in a.vertices.shape

    def test_wire(self, a: optika.apertures.AbstractAperture):
        wire = a.wire()
        assert isinstance(wire, na.AbstractCartesian3dVectorArray)
        assert "wire" in wire.shape
        assert wire.shape["wire"] == a.samples_wire

    class TestPlot(
        test_mixins.AbstractTestPlottable.TestPlot,
    ):
        def test_plot(
            self,
            a: optika.apertures.AbstractAperture,
            ax: None | matplotlib.axes.Axes | na.ScalarArray,
            transformation: None | na.transformations.AbstractTransformation,
        ):
            if na.unit_normalized(a.wire()).is_equivalent(u.mm):
                super().test_plot(
                    a=a,
                    ax=ax,
                    transformation=transformation,
                )


radius_parameterization = [
    0.5,
    100 * u.mm,
    na.linspace(100, 200, axis="radius", num=4) * u.mm,
    na.NormalUncertainScalarArray(100 * u.mm, width=10 * u.mm),
]


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.apertures.CircularAperture(
            radius=radius,
            samples_wire=21,
            active=active,
            inverted=inverted,
            transformation=transformation,
            kwargs_plot=kwargs_plot,
        )
        for radius in radius_parameterization
        for active in active_parameterization
        for inverted in inverted_parameterization
        for transformation in transform_parameterization
        for kwargs_plot in test_mixins.kwargs_plot_parameterization
    ],
)
class TestCircularAperture(
    AbstractTestAbstractAperture,
):
    def test_radius(self, a: optika.apertures.CircularAperture):
        assert isinstance(a.radius, (float, u.Quantity, na.AbstractScalar))
        assert np.all(a.radius >= 0)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.apertures.CircularSectorAperture(
            radius=radius,
            samples_wire=21,
            active=active,
            inverted=inverted,
            transformation=transformation,
            kwargs_plot=kwargs_plot,
        )
        for radius in radius_parameterization
        for active in active_parameterization
        for inverted in inverted_parameterization
        for transformation in transform_parameterization
        for kwargs_plot in test_mixins.kwargs_plot_parameterization
    ],
)
class TestCircularSectorAperture(
    AbstractTestAbstractAperture,
):
    def test_radius(self, a: optika.apertures.CircularSectorAperture):
        assert isinstance(a.radius, (float, u.Quantity, na.AbstractScalar))
        assert np.all(a.radius >= 0)

    def test_angle_start(self, a: optika.apertures.CircularSectorAperture):
        assert isinstance(a.radius, (float, u.Quantity, na.AbstractScalar))
        assert np.all(a.radius >= 0)

    def test_angle_stop(self, a: optika.apertures.CircularSectorAperture):
        assert isinstance(a.radius, (float, u.Quantity, na.AbstractScalar))
        assert np.all(a.radius >= 0)


class AbstractTestAbstractPolygonalAperture(
    AbstractTestAbstractAperture,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.apertures.EllipticalAperture(
            radius=radius,
            samples_wire=21,
            active=active,
            inverted=inverted,
            transformation=transformation,
            kwargs_plot=kwargs_plot,
        )
        for radius in [na.Cartesian2dVectorArray(50, 100) * u.mm]
        for active in active_parameterization
        for inverted in inverted_parameterization
        for transformation in transform_parameterization
        for kwargs_plot in test_mixins.kwargs_plot_parameterization
    ],
)
class TestEllipticalAperture(
    AbstractTestAbstractAperture,
):
    def test_radius(self, a: optika.apertures.EllipticalAperture):
        assert isinstance(a.radius, na.AbstractCartesian2dVectorArray)
        assert np.all(a.radius >= 0)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.apertures.PolygonalAperture(
            vertices=na.Cartesian3dVectorArray(
                x=na.ScalarArray([-1, 1, 1, -1] * u.mm, axes="vertex"),
                y=na.ScalarArray([-1, -1, 1, 1] * u.mm, axes="vertex"),
                z=0 * u.mm,
            ),
            samples_wire=21,
            active=active,
            inverted=inverted,
            transformation=transformation,
            kwargs_plot=kwargs_plot,
        )
        for active in active_parameterization
        for inverted in inverted_parameterization
        for transformation in transform_parameterization
        for kwargs_plot in test_mixins.kwargs_plot_parameterization
    ],
)
class TestPolygonalAperture(
    AbstractTestAbstractPolygonalAperture,
):
    pass


half_width_parameterization = [
    0.5,
    100 * u.mm,
    na.linspace(100, 200, axis="radius", num=4) * u.mm,
    na.NormalUncertainScalarArray(100 * u.mm, width=10 * u.mm),
    na.Cartesian2dVectorArray(100 * u.mm, 200 * u.mm),
]


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.apertures.RectangularAperture(
            half_width=half_width,
            samples_wire=21,
            active=active,
            inverted=inverted,
            transformation=transformation,
            kwargs_plot=kwargs_plot,
        )
        for half_width in half_width_parameterization
        for active in active_parameterization
        for inverted in inverted_parameterization
        for transformation in transform_parameterization
        for kwargs_plot in test_mixins.kwargs_plot_parameterization
    ],
)
class TestRectangularAperture(
    AbstractTestAbstractPolygonalAperture,
):
    def test_half_width(self, a: optika.apertures.RectangularAperture):
        types_valid = (
            float,
            u.Quantity,
            na.AbstractScalar,
            na.AbstractCartesian2dVectorArray,
        )
        assert isinstance(a.half_width, types_valid)
        assert np.all(a.half_width >= 0)


class AbstractTestAbstractRegularPolygonalAperture(
    AbstractTestAbstractPolygonalAperture,
):
    def test_radius(self, a: optika.apertures.AbstractRegularPolygonalAperture):
        assert isinstance(na.as_named_array(a.radius), na.AbstractScalar)

    def test_num_vertices(self, a: optika.apertures.AbstractRegularPolygonalAperture):
        assert isinstance(a.num_vertices, int)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.apertures.RegularPolygonalAperture(
            radius=radius,
            num_vertices=6,
            samples_wire=21,
            active=active,
            inverted=inverted,
            transformation=transformation,
            kwargs_plot=kwargs_plot,
        )
        for radius in radius_parameterization
        for active in active_parameterization
        for inverted in inverted_parameterization
        for transformation in transform_parameterization
        for kwargs_plot in test_mixins.kwargs_plot_parameterization
    ],
)
class TestRegularPolygonalAperture(
    AbstractTestAbstractRegularPolygonalAperture,
):
    pass


class AbstractTestAbstractOctagonalAperture(
    AbstractTestAbstractRegularPolygonalAperture,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.apertures.OctagonalAperture(
            radius=radius,
            samples_wire=21,
            active=active,
            inverted=inverted,
            transformation=transformation,
            kwargs_plot=kwargs_plot,
        )
        for radius in radius_parameterization
        for active in active_parameterization
        for inverted in inverted_parameterization
        for transformation in transform_parameterization
        for kwargs_plot in test_mixins.kwargs_plot_parameterization
    ],
)
class TestOctagonalAperture(
    AbstractTestAbstractOctagonalAperture,
):
    pass


class AbstractTestAbstractIsoscelesTrapezoidalAperture(
    AbstractTestAbstractPolygonalAperture,
):
    def test_x_left(self, a: optika.apertures.AbstractIsoscelesTrapezoidalAperture):
        assert isinstance(na.as_named_array(a.x_left), na.AbstractScalar)

    def test_x_right(self, a: optika.apertures.AbstractIsoscelesTrapezoidalAperture):
        assert isinstance(na.as_named_array(a.x_right), na.AbstractScalar)

    def test_angle(self, a: optika.apertures.AbstractIsoscelesTrapezoidalAperture):
        assert isinstance(na.as_named_array(a.angle), na.AbstractScalar)


x_left_parameterization = [
    25 * u.mm,
    na.linspace(25, 30, axis="x_left", num=2) * u.mm,
]


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.apertures.IsoscelesTrapezoidalAperture(
            x_left=x_left,
            x_right=50 * u.mm,
            angle=45 * u.deg,
            samples_wire=21,
            active=active,
            inverted=inverted,
            transformation=transformation,
            kwargs_plot=kwargs_plot,
        )
        for x_left in x_left_parameterization
        for active in active_parameterization
        for inverted in inverted_parameterization
        for transformation in transform_parameterization
        for kwargs_plot in test_mixins.kwargs_plot_parameterization
    ],
)
class TestIsoscelesTrapezoidalAperture(
    AbstractTestAbstractIsoscelesTrapezoidalAperture,
):
    pass



"""
Models of light sensors that can be used in optical systems.
"""

from typing import TypeVar, Sequence
import abc
import dataclasses
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
from . import AbstractImagingSensorMaterial, IdealImagingSensorMaterial

__all__ = [
    "AbstractImagingSensor",
    "ImagingSensor",
    "AbstractCCD",
]


MaterialT = TypeVar("MaterialT", bound=AbstractImagingSensorMaterial)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImagingSensor(
    optika.surfaces.AbstractSurface[
        None,
        MaterialT,
        optika.apertures.RectangularAperture,
        optika.apertures.RectangularAperture,
        None,
    ],
):
    """
    An interface describing an imaging sensor that can be used as the last
    surface in an optical system.
    """

    @property
    def sag(self) -> optika.sags.AbstractSag:
        return optika.sags.NoSag()

    @property
    def rulings(self) -> None:
        return None

    @property
    @abc.abstractmethod
    def width_pixel(self) -> u.Quantity | na.AbstractCartesian2dVectorArray:
        """
        The physical size of each pixel on the sensor.
        """

    @property
    @abc.abstractmethod
    def axis_pixel(self) -> na.Cartesian2dVectorArray[str, str]:
        """
        The names of the logical axes corresponding to the rows and
        columns of the pixel grid.
        """

    @property
    @abc.abstractmethod
    def num_pixel(self) -> na.Cartesian2dVectorArray[int, int]:
        """
        The number of pixels along each axis of the sensor.
        """

    @property
    @abc.abstractmethod
    def timedelta_exposure(self) -> u.Quantity | na.AbstractScalar:
        """
        The exposure time of the sensor.
        """

    @property
    def aperture(self):
        """
        The light-sensitive aperture of the sensor.
        """
        return optika.apertures.RectangularAperture(
            half_width=self.width_pixel * self.num_pixel / 2,
        )

    def readout(
        self,
        rays: optika.rays.RayVectorArray,
        wavelength: na.AbstractScalar,
        timedelta: None | u.Quantity | na.AbstractScalar = None,
        axis: None | str | Sequence[str] = None,
        where: bool | na.AbstractScalar = True,
    ) -> na.FunctionArray[
        na.SpectralPositionalVectorArray,
        na.AbstractScalar,
    ]:
        """
        Given a set of rays incident on the sensor surface,
        where each ray represents an expected number of photons per unit time,
        simulate the number of electrons that would be measured by the sensor.

        Parameters
        ----------
        rays
            A set of incident rays in local coordinates to measure.
        wavelength
            The edges of the wavelength bins to sample.
        timedelta
            The exposure time of the measurement.
            If :obj:`None` (the default), the value in :attr:`timedelta_exposure`
            will be used.
        axis
            The logical axes along which to collect photons.
        where
            A boolean mask used to indicate which photons should be considered
            when calculating the signal measured by the sensor.
        """
        if timedelta is None:
            timedelta = self.timedelta_exposure

        where = where & rays.unvignetted

        sgn = np.sign(rays.intensity)

        rays = dataclasses.replace(
            rays,
            intensity=np.abs(rays.intensity) * timedelta,
        )

        normal = self.sag.normal(rays.position)

        rays = self.material.signal(
            rays=rays,
            normal=normal,
        )

        rays = dataclasses.replace(
            rays,
            intensity=sgn * rays.intensity,
        )

        rays = self.material.charge_diffusion(
            rays=rays,
            normal=normal,
        )

        bins = na.SpectralPositionalVectorArray(
            wavelength=wavelength,
            position=na.Cartesian2dVectorLinearSpace(
                start=self.aperture.bound_lower.xy,
                stop=self.aperture.bound_upper.xy,
                axis=self.axis_pixel,
                num=self.num_pixel + 1,
            ),
        )

        return na.histogram(
            a=na.SpectralPositionalVectorArray(
                wavelength=rays.wavelength,
                position=rays.position.xy,
            ),
            bins=bins,
            axis=axis,
            weights=rays.intensity * where,
        )


@dataclasses.dataclass(eq=False, repr=False)
class ImagingSensor(
    AbstractImagingSensor,
):
    """
    An arbitrary imaging sensor described by a pixel grid and a light-sensitive
    material.
    """

    name: None | str = None
    """The human-readable name of this sensor."""

    width_pixel: u.Quantity | na.AbstractCartesian2dVectorArray = 0 * u.um
    """The physical size of each pixel on the sensor."""

    axis_pixel: na.Cartesian2dVectorArray[str, str] = None
    """
    The names of the logical axes corresponding to the rows and 
    columns of the pixel grid.
    """

    num_pixel: na.Cartesian2dVectorArray[int, int] = None
    """The number of pixels along each axis of the sensor."""

    timedelta_exposure: u.Quantity | na.AbstractScalar = 0 * u.s
    """The exposure time of the sensor."""

    material: AbstractImagingSensorMaterial = None
    """
    A model of the light-sensitive material composing this sensor.
    
    If :obj:`None` (the default), :class:`optika.sensors.IdealImagingSensor`
    will be used.
    """

    aperture_mechanical: optika.apertures.RectangularAperture = None
    """The shape of the physical substrate supporting the sensor."""

    is_field_stop: bool = False
    """A flag controlling whether this sensor is the field stop for the system."""

    is_pupil_stop: bool = False
    """A flag controlling whether this sensor is the pupil stop for the system."""

    transformation: None | na.transformations.AbstractTransformation = None
    """The position and orientation of the sensor in the global coordinate system."""

    kwargs_plot: None | dict = None
    """Extra keyword arguments to pass to :meth:`plot`"""

    def __post_init__(self) -> None:
        if self.material is None:
            self.material = IdealImagingSensorMaterial()

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.name),
            optika.shape(self.width_pixel),
            optika.shape(self.num_pixel),
            optika.shape(self.timedelta_exposure),
            optika.shape(self.transformation),
        )


class AbstractCCD(
    AbstractImagingSensor[MaterialT],
):
    pass



import numpy as np
import pytest
import astropy.units as u  # type: ignore[import]
import named_arrays as na  # type: ignore[import]
import optika
from . import test_mixins


class AbstractTestAbstractSag(
    test_mixins.AbstractTestPrintable,
    test_mixins.AbstractTestTransformable,
    test_mixins.AbstractTestShaped,
):
    def test_parameters_slope_error(self, a: optika.sags.AbstractSag):
        if a.parameters_slope_error is not None:
            assert isinstance(
                a.parameters_slope_error,
                optika.metrology.SlopeErrorParameters,
            )

    def test_parameters_roughness(self, a: optika.sags.AbstractSag):
        if a.parameters_roughness is not None:
            assert isinstance(
                a.parameters_roughness,
                optika.metrology.RoughnessParameters,
            )

    def test_parameters_microroughness(self, a: optika.sags.AbstractSag):
        if a.parameters_microroughness is not None:
            assert isinstance(
                a.parameters_microroughness,
                optika.metrology.RoughnessParameters,
            )

    @pytest.mark.parametrize(
        argnames="position",
        argvalues=[
            na.Cartesian3dVectorArray() * u.mm,
            na.Cartesian3dVectorLinearSpace(0, 1, axis="s", num=5) * u.mm,
            na.Cartesian3dVectorArray(
                x=na.ScalarLinearSpace(0, 1, axis="x", num=5) * u.mm,
                y=na.NormalUncertainScalarArray(
                    nominal=na.ScalarLinearSpace(-1, 0, axis="y", num=6) * u.mm,
                    width=0.1 * u.mm,
                ),
                z=0 * u.mm,
            ),
        ],
    )
    class TestFunctionsOfPosition:
        def test__call__(
            self,
            a: optika.sags.AbstractSag,
            position: na.AbstractCartesian3dVectorArray,
        ):
            result = a(position)
            assert isinstance(na.as_named_array(result), na.AbstractScalar)
            if na.shape(result):
                assert set(na.shape(position)).issubset(na.shape(result))

        def test_normal(
            self,
            a: optika.sags.AbstractSag,
            position: na.AbstractCartesian3dVectorArray,
        ):
            result = a.normal(position)
            assert isinstance(result, na.AbstractCartesian3dVectorArray)
            if na.shape(result):
                assert set(na.shape(position)).issubset(na.shape(result))

    @pytest.mark.parametrize(
        argnames="rays",
        argvalues=[
            optika.rays.RayVectorArray(
                wavelength=500 * u.nm,
                position=na.Cartesian3dVectorArray(1, 2, 3) * u.mm,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
            optika.rays.RayVectorArray(
                wavelength=na.NormalUncertainScalarArray(500 * u.nm, width=5 * u.nm),
                position=na.Cartesian3dVectorArray(
                    x=na.NormalUncertainScalarArray(1 * u.mm, width=0.1 * u.mm),
                    y=na.NormalUncertainScalarArray(2 * u.mm, width=0.1 * u.mm),
                    z=na.NormalUncertainScalarArray(3 * u.mm, width=0.1 * u.mm),
                ),
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
            optika.rays.RayVectorArray(
                wavelength=na.linspace(500, 600, axis="y", num=5) * u.nm,
                position=na.Cartesian3dVectorLinearSpace(0, 5, axis="y", num=5) * u.mm,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
        ],
    )
    class TestFunctionsOfRays:
        def test_intercept(
            self,
            a: optika.sags.AbstractSag,
            rays: optika.rays.AbstractRayVectorArray,
        ):
            result = a.intercept(rays)
            assert isinstance(result, optika.rays.AbstractRayVectorArray)
            assert np.all(np.isfinite(result.position))
            assert np.allclose(a(result.position), result.position.z)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.sags.NoSag(
            parameters_slope_error=optika.metrology.SlopeErrorParameters(
                kernel_size=2 * u.mm,
                step_size=4 * u.mm,
            ),
            parameters_roughness=optika.metrology.RoughnessParameters(
                period_min=2 * u.mm,
                period_max=4 * u.mm,
            ),
            parameters_microroughness=optika.metrology.RoughnessParameters(
                period_min=0.1 * u.mm,
                period_max=2 * u.mm,
            ),
        )
    ],
)
class TestNoSag(
    AbstractTestAbstractSag,
):
    pass


class AbstractTestAbstractSphericalSag(
    AbstractTestAbstractSag,
):
    def test_curvature(self, a: optika.sags.SphericalSag):
        assert isinstance(na.as_named_array(a.curvature), na.AbstractScalar)
        assert na.shape(a.curvature) == na.shape(a.radius)


def radius_parameterization() -> list[u.Quantity | na.AbstractScalar]:
    nominals = [
        100 * u.mm,
        na.ScalarLinearSpace(100, 1000, axis="radius", num=5) * u.mm,
    ]
    widths = [
        None,
        10 * u.mm,
    ]
    return [
        nominal if width is None else na.NormalUncertainScalarArray(nominal, width)
        for nominal in nominals
        for width in widths
    ]


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.sags.SphericalSag(
            radius=radius,
            transformation=transformation,
        )
        for radius in radius_parameterization()
        for transformation in test_mixins.transformation_parameterization
    ],
)
class TestSphericalSag(
    AbstractTestAbstractSphericalSag,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.sags.CylindricalSag(
            radius=radius,
            transformation=transformation,
        )
        for radius in radius_parameterization()
        for transformation in test_mixins.transformation_parameterization
    ],
)
class TestCylindricalSag(
    AbstractTestAbstractSag,
):
    pass


class AbstractTestAbstractConicSag(
    AbstractTestAbstractSag,
):
    def test_radius(self, a: optika.sags):
        assert isinstance(na.as_named_array(a.radius), na.ScalarLike)
        assert na.unit_normalized(a.radius).is_equivalent(u.mm)

    def test_conic(self, a: optika.sags):
        assert isinstance(na.as_named_array(a.conic), na.ScalarLike)
        assert na.unit_normalized(a.conic).is_equivalent(u.dimensionless_unscaled)


def conic_parameterization() -> list[u.Quantity | na.AbstractScalar]:
    nominals = [
        0 * u.dimensionless_unscaled,
        na.ScalarLinearSpace(0, 1, axis="conic", num=4),
    ]
    widths = [
        None,
        0.1 * u.dimensionless_unscaled,
    ]
    return [
        nominal if width is None else na.NormalUncertainScalarArray(nominal, width)
        for nominal in nominals
        for width in widths
    ]


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.sags.ConicSag(
            radius=radius,
            conic=conic,
            transformation=transformation,
        )
        for radius in radius_parameterization()
        for conic in conic_parameterization()
        for transformation in test_mixins.transformation_parameterization
    ],
)
class TestConicSag(
    AbstractTestAbstractConicSag,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.sags.ParabolicSag(
            focal_length=radius / 2,
            transformation=transformation,
        )
        for radius in radius_parameterization()
        for transformation in test_mixins.transformation_parameterization
    ],
)
class TestParabolicSag(
    AbstractTestAbstractConicSag,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.sags.ToroidalSag(
            radius=radius,
            radius_of_rotation=2 * radius_of_rotation,
            transformation=transformation,
        )
        for radius in radius_parameterization()
        for radius_of_rotation in radius_parameterization()
        for transformation in test_mixins.transformation_parameterization
    ],
)
class TestToroidalSag(
    AbstractTestAbstractSphericalSag,
):
    pass



from ._vectors_polarization import (
    AbstractPolarizationVectorArray,
    PolarizationVectorArray,
)
from ._vectors_field import (
    AbstractFieldVectorArray,
    FieldVectorArray,
)
from ._vectors_pupil import (
    AbstractPupilVectorArray,
    PupilVectorArray,
)
from ._vectors_scene import (
    AbstractSceneVectorArray,
    SceneVectorArray,
)
from ._vectors_object import (
    AbstractObjectVectorArray,
    ObjectVectorArray,
)

__all__ = [
    "AbstractPolarizationVectorArray",
    "PolarizationVectorArray",
    "AbstractFieldVectorArray",
    "FieldVectorArray",
    "AbstractPupilVectorArray",
    "PupilVectorArray",
    "AbstractSceneVectorArray",
    "SceneVectorArray",
    "AbstractObjectVectorArray",
    "ObjectVectorArray",
]



from __future__ import annotations
from typing import TypeVar
from typing_extensions import Self
import dataclasses
import named_arrays as na
from . import AbstractFieldVectorArray, FieldVectorArray

__all__ = [
    "AbstractSceneVectorArray",
    "SceneVectorArray",
]


WavelengthT = TypeVar("WavelengthT", bound=na.ScalarLike)
FieldT = TypeVar("FieldT", bound=na.AbstractCartesian2dVectorArray)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractSceneVectorArray(
    AbstractFieldVectorArray,
    na.AbstractSpectralVectorArray,
):
    @property
    def type_abstract(self) -> type[AbstractSceneVectorArray]:
        return AbstractSceneVectorArray

    @property
    def type_explicit(self) -> type[SceneVectorArray]:
        return SceneVectorArray

    @property
    def type_matrix(self) -> type[na.AbstractMatrixArray]:
        raise NotImplementedError


@dataclasses.dataclass(eq=False, repr=False)
class SceneVectorArray(
    AbstractSceneVectorArray,
    FieldVectorArray[FieldT],
    na.SpectralVectorArray[WavelengthT],
):
    @classmethod
    def from_scalar(
        cls,
        scalar: na.AbstractScalar,
        like: None | Self = None,
    ) -> Self:
        if like is not None:
            return type(like)(
                wavelength=scalar,
                field=scalar,
                # field=na.Cartesian2dVectorArray.from_scalar(scalar, like=like.field)
            )
        else:
            return cls(
                wavelength=scalar,
                field=scalar,
                # field=na.Cartesian2dVectorArray.from_scalar(scalar),
            )



"""
A Python package for simulating and designing optical systems.
"""

from . import mixins
from ._util import shape, direction, angles
from . import vectors
from . import targets
from . import rays
from . import metrology
from . import sags
from . import chemicals
from . import materials
from . import apertures
from . import rulings
from . import propagators
from . import surfaces
from . import sensors
from . import systems

__all__ = [
    "mixins",
    "shape",
    "direction",
    "angles",
    "vectors",
    "targets",
    "rays",
    "metrology",
    "sags",
    "chemicals",
    "materials",
    "apertures",
    "rulings",
    "propagators",
    "surfaces",
    "sensors",
    "systems",
]



from ._ray_vectors import (
    AbstractRayVectorArray,
    RayVectorArray,
)
from ._ray_functions import (
    AbstractRayFunctionArray,
    RayFunctionArray,
)

__all__ = [
    "AbstractRayVectorArray",
    "RayVectorArray",
    "AbstractRayFunctionArray",
    "RayFunctionArray",
]



import pytest
import astropy.units as u
import optika
from . import test_multilayers


class AbstractTestAbstractThinFilmFilter(
    test_multilayers.AbstractTestAbstractMultilayerFilm,
):

    def test_layer(self, a: optika.materials.AbstractThinFilmFilter):
        result = a.layer
        assert isinstance(result, optika.materials.AbstractLayer)

    def test_layer_oxide(self, a: optika.materials.AbstractThinFilmFilter):
        result = a.layer_oxide
        assert isinstance(result, optika.materials.AbstractLayer)

    def test_mesh(self, a: optika.materials.AbstractThinFilmFilter):
        result = a.mesh
        assert isinstance(result, optika.materials.meshes.AbstractMesh)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.materials.ThinFilmFilter(
            layer=optika.materials.Layer(
                chemical="Al",
                thickness=100 * u.nm,
            ),
            layer_oxide=optika.materials.Layer(
                chemical="Al2O3",
                thickness=2 * u.nm,
            ),
            mesh=optika.materials.meshes.Mesh(
                chemical="Ni",
                efficiency=0.8,
                pitch=70 / u.mm,
            ),
        )
    ],
)
class TestThinFilmFilter(
    AbstractTestAbstractThinFilmFilter,
):
    pass



import pathlib
import matplotlib.pyplot as plt
import pymupdf
import named_arrays as na

__all__ = [
    "airforce",
]


def airforce(
    axis_x: str,
    axis_y: str,
    num_x: int = 1000,
    num_y: int = 1000,
) -> na.ScalarArray:
    """
    A `1951 USAF resolution test target <https://en.wikipedia.org/wiki/1951_USAF_resolution_test_chart>`_
    used for testing the performance of optical systems.

    Parameters
    ----------
    axis_x:
        The name of the horizontal axis.
    axis_y:
        The name of the vertical axis.
    num_x:
        The number of pixels along the horizontal axis.
    num_y:
        The number of pixels along the vertical axis.

    Examples
    --------

    Load and display the test target

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import named_arrays as na
        import optika

        # Load the test target
        target = optika.targets.airforce("x", "y")

        # Display the test target
        fig, ax = plt.subplots(constrained_layout=True)
        na.plt.pcolormesh(C=target);
    """
    path = pathlib.Path(__file__).parent / "USAF-1951.svg"

    doc = pymupdf.open(path)

    page = doc[0]

    zoom_x = num_x / page.rect.width
    zoom_y = num_y / page.rect.height

    mat = pymupdf.Matrix(zoom_x, zoom_y)

    pix = page.get_pixmap(matrix=mat)

    fn = "tmp.png"
    pix.save(fn)

    img = plt.imread(fn)

    img = img.astype(float)[::-1].sum(~0)

    img = img / img.max()

    img = -img + 1

    return na.ScalarArray(
        ndarray=img,
        axes=(axis_y, axis_x),
    )
